#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/icmp.h>
#include <linux/udp.h>
#include <linux/gfp.h>
#include <linux/ip.h>
#include <linux/ktime.h>
#include <linux/time.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <net/ip.h>
#include <net/checksum.h>
#include <linux/spinlock.h>

#include <linux/sysctl.h>
#include <linux/version.h>

enum
{
	FILTER_DF_ONLY_FLAG_OFF	= 0,
	FILTER_DF_ONLY_FLAG_ON,
	FILTER_DF_ALL,
};

#define OLD_KERNEL_HOOK_5_0 (LINUX_VERSION_CODE < KERNEL_VERSION(5,0,0))
#define OLD_KERNEL_HOOK_4_18 (LINUX_VERSION_CODE <= KERNEL_VERSION(4,18,0))
#define OLD_KERNEL_HOOK_4_13 (LINUX_VERSION_CODE <= KERNEL_VERSION(4,13,0))
#define OLD_KERNEL_HOOK_4_4 (LINUX_VERSION_CODE <= KERNEL_VERSION(4,4,0))
#define OLD_KERNEL_HOOK_4 (LINUX_VERSION_CODE <= KERNEL_VERSION(4,1,0))
#define OLD_KERNEL_HOOK_3 (LINUX_VERSION_CODE <= KERNEL_VERSION(3,12,0))
#define OLD_KERNEL_SYSCTL (LINUX_VERSION_CODE <= KERNEL_VERSION(3,0,0))
#define OLD_KERNEL (LINUX_VERSION_CODE <= KERNEL_VERSION(3,0,0))


#define DDD_UDP_LICENSE "GPL"
#define DDD_UDP_AUTHOR	"q2l <@.>"
#define DDD_UDP_DESCRIPTION	"Dummy Drop Duplicate UDP/TCP filter"
#define DDD_UDP_VERSION	"0.7.3"

#undef DEBUG

#ifdef DEBUG
#define local_debug(...) \
        do {pr_info(__VA_ARGS__);} while(0)
#else
#define local_debug(...) \
        while(0) {pr_info(__VA_ARGS__);}
#endif


#define IFNAME_SIZE			16

struct ip_packet
{
	/* from sk_buff */
	struct timeval			tstamp;
	/* from iphdr */
	long int   				id;
};

struct ddd_peer
{
	/* from iphdr */ 	
	__u32   saddr;
	__u32   daddr;
	/* from udphdr / tcphdr */
	__be16  source;
	__be16  dest;
	
	unsigned long		total_pkts;
	unsigned long		dup_pkts;
	
	struct ip_packet	*history;
	int					history_size;
	int					curr_history_pos;
	int					inuse;
	int					filterable;
};

struct ddd_filter {

	unsigned long				total_pkts;
	unsigned long				dup_pkts;

	int							enable;
	spinlock_t					lock;

	struct ddd_peer				*cache;
	int							cache_curr_size;
	int							cache_size;
	
	int							cache_curr_max_size;
	int 						cache_max_size;

	int							history_curr_max_size;
	int							history_max_size;
	int							ip_packet_ttl;
	int							peer_ttl;
	int							filter_df;
	char 						ifname[IFNAME_SIZE];
	
	struct ctl_table 			*table;
	struct ctl_table_header 	*table_header;
	
#if OLD_KERNEL_SYSCTL
	struct ctl_path 			ddd_udp_ctl_path[5];
#endif

	char						ddd_stat[8192];
	struct timeval				tstamp_ddd_stat;
	int							update_ddd_stat_period;
	int							update_ddd_force;
};


static struct ddd_filter	ddd_filters[1] =
{
	{
		.update_ddd_force		= 0,
		.update_ddd_stat_period = 10000,
		.ddd_stat				= {0},
		.total_pkts				= 0,
		.dup_pkts				= 0,

		.enable					= 0,
		.filter_df				= FILTER_DF_ONLY_FLAG_OFF,

		.cache_curr_max_size	= 128,
		.cache_max_size			= 128,

		.history_curr_max_size	= 8,
		.history_max_size		= 8,
		
		.ip_packet_ttl			= 1000,
		.peer_ttl				= 60000,
		.ifname 				= {0},
		
#if OLD_KERNEL_SYSCTL
 		.ddd_udp_ctl_path 		= {
			{ .procname = "net", },
			{ .procname = "ipv4", },
			{ .procname = "ddd_udp", },
			{ .procname = "0", },
			{ }
		},
#endif
		.table					= NULL,
		.table_header			= NULL		
	}
};


static void cache_delete_all(struct ddd_filter *filter);
static int get_peer_last_touch(struct ddd_filter *filter, struct ddd_peer *peer, struct timeval	*ptstamp);
static long tvdiff(struct timeval *newer, struct timeval *older);
static int tvcmp(const struct timeval *lhs, const struct timeval *rhs);
static int update_stat_ddd_filter(struct ddd_filter	*filter, int force);

/* -------------------------------------------------------------------------- */
static void ddd_gettimeofday(struct timeval *tv)
{
#if OLD_KERNEL_HOOK_5_0
	if (tv)
	{
		do_gettimeofday(tv);
	}
#else
	if (tv)
	{
		struct timespec64 ts;
		ktime_get_real_ts64(&ts);
		tv->tv_sec = ts.tv_sec;
		tv->tv_usec = ts.tv_nsec/1000;
	}
#endif
}
/* -------------------------------------------------------------------------- */
static struct ddd_filter	*find_ddd_filter(struct ddd_filter	*filters, int filters_count, char *ifname)
{
	int i;
	struct ddd_filter	*filter = NULL;
	
	if (filters && ifname)
	{
		for (i=0;(i<filters_count) && (!filter);i++)
		{
			if ((ifname[0] != 0) && (filters[i].ifname[0] != 0) && 
				(strcmp(filters->ifname, ifname) == 0))
				filter = &filters[i];
		}
	}
	
	return filter;
}
/* -------------------------------------------------------------------------- */
static int close_ddd_udp(struct ddd_filter	*filter)
{
	if (filter)
	{
		filter->enable = 0;
		cache_delete_all(filter);
	}
	return 0;
}
/* -------------------------------------------------------------------------- */
static int init_ddd_udp(struct ddd_filter	*filter)
{
	int i;
	struct ddd_peer 	*peer=NULL;
	if (filter)
	{
		spin_lock_init(&filter->lock);
		
		filter->enable				= 0;
		filter->cache_max_size 		= 128;
		filter->cache_curr_max_size	= 128;
		filter->cache_size			= 128;
		filter->cache_curr_size		= 0;		
		filter->cache				= kmalloc(filter->cache_size * sizeof(struct ddd_peer), GFP_ATOMIC);
		
		filter->total_pkts 				= 0;
		filter->dup_pkts 				= 0;
		filter->filter_df		 		= FILTER_DF_ALL;
		filter->history_max_size 		= 8;
		filter->history_curr_max_size 	= 8;
		filter->ip_packet_ttl 			= 1000; /* us */
		filter->peer_ttl 				= 60000; /* us */
		memset(filter->ifname, 0, sizeof(filter->ifname));
		
		peer = filter->cache;
		for (i=0;i<filter->cache_size;i++,peer++)
		{
			peer->inuse			= 0;
			peer->filterable	= 0;
			peer->history_size 	= filter->history_curr_max_size;
			peer->history 		= kmalloc(peer->history_size * sizeof(struct ip_packet), GFP_ATOMIC);
		}
		filter->update_ddd_stat_period	= 10000;
		ddd_gettimeofday(&filter->tstamp_ddd_stat);

		update_stat_ddd_filter(filter, 1);
	}
	return 0;
}
/* -------------------------------------------------------------------------- */
static char persent_char[]="per.";
static int update_stat_ddd_filter(struct ddd_filter	*filter, int force)
{
	int i,j;
	struct ddd_peer 	*peer=NULL;
	unsigned long 		persent, persent_fract, valx100;
	char				src_str[64], dst_str[64];	
	struct timeval 		now_stamp, peer_stamp;
	unsigned long		curr_size, stat_buffer_size, trust_size;
	char				*pstat_buffer;
	long				update_time, peer_time;

	curr_size = 0;

	ddd_gettimeofday(&now_stamp);

	if (filter)
	{
		update_time = tvdiff(&now_stamp, &filter->tstamp_ddd_stat);
		if ((update_time < filter->update_ddd_stat_period) && !force)
		{
			return 0;
		}
		
		filter->tstamp_ddd_stat = now_stamp;
		
		stat_buffer_size = sizeof(filter->ddd_stat)-1;
		trust_size = (stat_buffer_size>>1)+(stat_buffer_size>>2);
		memset(filter->ddd_stat, 0, sizeof(filter->ddd_stat));
		pstat_buffer = filter->ddd_stat;
		

		valx100 = 100 * filter->dup_pkts;
		persent = (filter->total_pkts == 0) ? 0 : (valx100 / filter->total_pkts);
		persent_fract = valx100 - persent * filter->total_pkts;
		persent_fract = (filter->total_pkts == 0) ? 0 : (100 * persent_fract / filter->total_pkts);	
		
		curr_size += snprintf(&pstat_buffer[curr_size], trust_size-curr_size,
					"Dummy Drop Duplicate UDP packets:\n" );
		if (curr_size > trust_size) return curr_size;
		curr_size += snprintf(&pstat_buffer[curr_size], trust_size-curr_size,
					"version\t\t\t: %s\n", DDD_UDP_VERSION);                     
		if (curr_size > trust_size) return curr_size;
		curr_size += snprintf(&pstat_buffer[curr_size], trust_size-curr_size,
					"enabled\t\t\t: %d\n", filter->enable); 
		if (curr_size > trust_size) return curr_size;
		curr_size += snprintf(&pstat_buffer[curr_size], trust_size-curr_size,
					"filter_df\t\t: %d\n", filter->filter_df);
		if (curr_size > trust_size) return curr_size;
		curr_size += snprintf(&pstat_buffer[curr_size], trust_size-curr_size,
					"ifname\t\t\t: %s\n", filter->ifname); 
		if (curr_size > trust_size) return curr_size;
		curr_size += snprintf(&pstat_buffer[curr_size], trust_size-curr_size,
					"total packets\t\t: %ld\n", filter->total_pkts); 
		if (curr_size > trust_size) return curr_size;
		curr_size += snprintf(&pstat_buffer[curr_size], trust_size-curr_size,
					"dup packets\t\t: %ld (%ld.%02ld %s)\n", filter->dup_pkts, persent, persent_fract, persent_char);
		if (curr_size > trust_size) return curr_size;
		curr_size += snprintf(&pstat_buffer[curr_size], trust_size-curr_size,
					"cache size\t\t: %d / %d\n", filter->cache_curr_size, filter->cache_size); 
		if (curr_size > trust_size) return curr_size;
		curr_size += snprintf(&pstat_buffer[curr_size], trust_size-curr_size,
					"history size\t\t: %d\n", filter->history_curr_max_size); 
		if (curr_size > trust_size) return curr_size;
		curr_size += snprintf(&pstat_buffer[curr_size], trust_size-curr_size,
					"packet ttl,(us)\t\t: %d\n", filter->ip_packet_ttl); 
		if (curr_size > trust_size) return curr_size;
		curr_size += snprintf(&pstat_buffer[curr_size], trust_size-curr_size,
					"peer ttl,(us)\t\t: %d\n", filter->peer_ttl); 
		if (curr_size > trust_size) return curr_size;
		curr_size += snprintf(&pstat_buffer[curr_size], trust_size-curr_size,
					"peer list\t\t:\n");
		if (curr_size > trust_size) return curr_size;
		curr_size += snprintf(&pstat_buffer[curr_size], trust_size-curr_size,
					"source                   dest                     total packets  dup packets    last time    bad packets\n");
		if (curr_size > trust_size) return curr_size;
		if (filter->cache && (filter->cache_curr_size > 0))
		{
			peer = filter->cache;
			j = filter->cache_curr_size;
			for (i=0;(i<filter->cache_size)&&(j>0);i++,peer++)
			{
				if (peer->inuse)
				{
					j--;
					valx100 = 100 * peer->dup_pkts;
					persent = (peer->total_pkts == 0) ? 0 : (valx100 / peer->total_pkts);
					persent_fract = valx100 - persent * peer->total_pkts;
					persent_fract = (peer->total_pkts == 0) ? 0 : (100 * persent_fract / peer->total_pkts);


					get_peer_last_touch(filter, peer, &peer_stamp);
					peer_time = tvdiff( &now_stamp, &peer_stamp);
					
					memset(src_str, 0, sizeof(src_str));
					snprintf(src_str, sizeof(src_str)-1, "%pI4:%d", &peer->saddr, ntohs(peer->source));
					memset(dst_str, 0, sizeof(dst_str));
					snprintf(dst_str, sizeof(dst_str)-1, "%pI4:%d", &peer->daddr, ntohs(peer->dest));
					curr_size += snprintf(&pstat_buffer[curr_size], trust_size-curr_size,
							"%-25s%-25s%-15ld%ld (%ld.%02ld %s)  %ld(%d)\n", 
							src_str, dst_str, peer->total_pkts, peer->dup_pkts, 
							persent, persent_fract, persent_char,
							peer_time, peer->curr_history_pos);
					if (curr_size > trust_size) return curr_size;							
				}
			}
		}		
	}
	return curr_size;
}
/* -------------------------------------------------------------------------- */
static struct ip_packet	*history_add(struct ddd_filter *filter, struct ddd_peer *peer, __u16   id,	struct timeval		*ptstamp)
{	
	struct ip_packet	*ipp = NULL;
	if ((filter != NULL) && (peer != NULL) && (ptstamp != NULL))
	{		
		if (peer->curr_history_pos < 0)
			peer->curr_history_pos = peer->history_size-1;
			
		ipp = &peer->history[peer->curr_history_pos];
		peer->curr_history_pos = peer->curr_history_pos-1;
		ipp->id = id;
		memcpy(&ipp->tstamp, ptstamp, sizeof(struct timeval));
		
		if (peer->curr_history_pos < 0)
			peer->curr_history_pos = peer->history_size-1;
	}
	
	return ipp;
}
/* -------------------------------------------------------------------------- */
static struct ip_packet  *history_find(struct ddd_filter *filter, struct ddd_peer *peer, __u16   id)
{
	struct ip_packet  	*ipp = NULL;
	int 				i,pos = 0;
	
	
	if (filter && peer)
	{
		if (peer->history && (peer->history_size > 0))
		{
			pos = peer->curr_history_pos + 1;
			if (pos >= peer->history_size) pos = 0;			
			ipp = &peer->history[pos];
			
			for (i=0;i<peer->history_size;i++,pos++,ipp++)
			{
				if (pos >= peer->history_size)
				{
					ipp = peer->history;
					pos = 0;
				}
			
				if (ipp->id == id) return ipp;
			}
		}
	}
	
	return NULL;
}
/* -------------------------------------------------------------------------- */
static void history_fill(struct ddd_filter *filter, struct ddd_peer *peer)
{
	struct ip_packet  	*ipp = NULL;
	int 				i;
	int 				delay_sec;
	
	
	if (filter && peer)
	{
		if (peer->history && (peer->history_size > 0))
		{
			ipp = peer->history;
			delay_sec = (filter->ip_packet_ttl/1000)+1;
			
			for (i=0;i<peer->history_size;i++,ipp++)
			{
				ipp->id = (-1);
				ddd_gettimeofday(&ipp->tstamp);
				ipp->tstamp.tv_sec -= delay_sec;
			}
		}
	}
}
/* -------------------------------------------------------------------------- */
static void history_delete_all(struct ddd_filter *filter, struct ddd_peer *peer)
{
    if (filter && peer)
    {
		history_fill(filter, peer);
	}
}
/* -------------------------------------------------------------------------- */
#ifdef DEBUG
static int peer_print(struct ddd_peer *peer, char *buffer, size_t buffer_size)
{
	size_t 				reuslt = 0;
	if (peer && buffer && (buffer_size>0))
	{
		memset(buffer, 0, buffer_size);
		reuslt += snprintf(&buffer[reuslt], buffer_size-reuslt, "%pI4:%d - %pI4:%d",
		
				&peer->saddr,
				ntohs(peer->source), 
				&peer->daddr,
				ntohs(peer->dest)
		);
	}
	return reuslt;
}
#endif
/* -------------------------------------------------------------------------- */
static struct ddd_peer *cache_find(struct ddd_filter *filter, __u32   saddr, __u32   daddr, __be16  source, __be16  dest)
{
		int 				i = 0, j = 0;
        struct ddd_peer 	*peer=NULL;
		
		if (filter)
		{			
			if (filter->cache && (filter->cache_curr_size > 0))
			{
				peer	= filter->cache;
				j 		= filter->cache_curr_size;
				/* find */
				for (i=0;(i<filter->cache_size)&&(j>0);i++,peer++)
				{
					if (peer->inuse)
					{
						j--;
						if ((peer->dest == dest) &&
								(peer->source == source) &&
								(peer->saddr == saddr) &&
								(peer->daddr == daddr))
						{
							return peer;
						}
					}
				}
			}
		}       
        return NULL;
}
/* -------------------------------------------------------------------------- */
static void cache_delete(struct ddd_filter *filter, int i)
{
	if (filter && (i<filter->cache_size) && (0<filter->cache_curr_size))
	{
		filter->cache_curr_size--;
		filter->cache[i].inuse = 0;
		filter->cache[i].filterable = 0;
	}
}
/* -------------------------------------------------------------------------- */
static void cache_delete_old(struct ddd_filter *filter)
{
    struct ddd_peer  	*peer;
    struct timeval 		now_stamp, peer_stamp;
    int 				i,j;
	long				peer_time;
    
	local_debug("cache_delete_old\n");
   
    if (filter)
    {
		if (filter->cache && (filter->cache_curr_size > 0))
		{
			ddd_gettimeofday(&now_stamp);
			peer	= filter->cache;
			j 		= filter->cache_curr_size;
			for (i=0;(i<filter->cache_size)&&(j>0);i++,peer++)
			{
				if (peer->inuse)
				{
					j--;
					get_peer_last_touch(filter, peer, &peer_stamp);
					peer_time = tvdiff( &now_stamp, &peer_stamp);

					if (peer_time > filter->peer_ttl)
					{				
						cache_delete(filter, i);
					}
				}
			}
		}
	}
}
/* -------------------------------------------------------------------------- */
static void cache_delete_all(struct ddd_filter *filter)
{
    struct ddd_peer  	*peer;
    int 				i;
    
    if (filter)
    {
		spin_lock(&filter->lock);
		
		peer = filter->cache;
		for (i=0;i<filter->cache_size;i++, peer++)		
		{
			if (peer->inuse)
			{
				history_delete_all(filter, peer);				
			}
			peer->inuse = 0;
			peer->filterable = 0;
		}
		filter->cache_curr_size = 0;		
		
		update_stat_ddd_filter(filter, 1);
		
		spin_unlock(&filter->lock);
	}
}
/* -------------------------------------------------------------------------- */
static int get_peer_last_touch(struct ddd_filter *filter, struct ddd_peer *peer, struct timeval	*ptstamp)
{
	int pos;
	struct ip_packet *ipp = NULL;
	if (filter && peer && ptstamp)
	{
		memset(ptstamp, 0, sizeof(struct timeval));

		pos = peer->curr_history_pos + 1;
		if (pos >= peer->history_size) pos = 0;
		ipp = &peer->history[pos];
		if (ipp)
		{
			memcpy(ptstamp, &ipp->tstamp, sizeof(struct timeval));
		}
		else
		{
			ddd_gettimeofday(ptstamp);
		}
	}
	return 0;
}
/* -------------------------------------------------------------------------- */
static long tvdiff(struct timeval *newer, struct timeval *older)
{
  return (newer->tv_sec-older->tv_sec)*1000+
    (newer->tv_usec-older->tv_usec)/1000;
}
/* -------------------------------------------------------------------------- */
static int tvcmp(const struct timeval *lhs, const struct timeval *rhs)
{
	if (lhs->tv_sec < rhs->tv_sec)
		return -1;
	if (lhs->tv_sec > rhs->tv_sec)
		return 1;
	return lhs->tv_usec - rhs->tv_usec;
}
/* -------------------------------------------------------------------------- */
static struct ddd_peer *cache_add(struct ddd_filter *filter, __u32   saddr, __u32   daddr, __be16  source, __be16  dest)
{
	int					i,j;
	struct ddd_peer 	*peer = NULL, *new_cache;
	struct timeval 		index_stamp, outcast_stamp;
	int 				outcast_peer;
#ifdef DEBUG		
	char				peer_buffer[256];
#endif	

	if (filter)
	{
		/* delete all old peer with big ttl */
       	cache_delete_old(filter);
       	/* realloc mem */
		if (filter->cache_size < filter->cache_curr_max_size)
		{
			new_cache	= kmalloc(filter->cache_curr_max_size * sizeof(struct ddd_peer), GFP_ATOMIC);
			if (new_cache)
			{
				peer = new_cache;					
				for (i=0;i<filter->cache_curr_max_size;i++,peer++)
				{
					peer->inuse			= 0;
					peer->filterable	= 0;
					peer->history_size	= filter->history_curr_max_size;
					peer->history 		= kmalloc(peer->history_size * sizeof(struct ip_packet), GFP_ATOMIC);
				}
				
				kfree(filter->cache);
				
				filter->cache = new_cache;
				filter->cache_size = filter->cache_curr_max_size;
			}
		}

       	/* delete oldest peer */		
        if (filter->cache_curr_size >= filter->cache_size)
        {
			outcast_peer = -1;
			peer	= filter->cache;
			j 		= filter->cache_curr_size;
			for (i=0;(i<filter->cache_size)&&(j>0);i++,peer++)
			{
				if (peer->inuse)
				{
					j--;
					if (outcast_peer < 0)
					{
						outcast_peer = i;
						get_peer_last_touch(filter, peer, &outcast_stamp);
					}
					else
					{
						get_peer_last_touch(filter, peer, &index_stamp);
						if (tvcmp(&index_stamp, &outcast_stamp) < 0)
						{
							outcast_peer	= i;
							outcast_stamp 	= index_stamp;
						}
					}
				}
			}

			local_debug("cache_delete\n");
			cache_delete(filter, outcast_peer);
        }
        
        /* get first unuse */
		peer	= filter->cache;
		j 		= -1;
		for (i=0;(i<filter->cache_size)&&(j<0);i++,peer++)
		{
			if (!peer->inuse)
			{
				j = i;
			}
		}
		/* bad idea */
        if (j<0) j = 0;
        
        filter->cache_curr_size++;
		peer = &filter->cache[j];
		peer->inuse = 1;
		
		peer->curr_history_pos	= 0;
		if (peer->history_size < filter->history_curr_max_size)
		{
			if (peer->history) kfree(peer->history);
			peer->history_size 		= filter->history_curr_max_size;
			peer->history			= kmalloc(peer->history_size * sizeof(struct ip_packet), GFP_ATOMIC);
		}
		
		history_fill(filter, peer);
		
		peer->saddr 			= saddr;
		peer->daddr 			= daddr;
		peer->source 			= source;
		peer->dest 				= dest;

		peer->total_pkts		= 0;
		peer->dup_pkts			= 0;
		
#ifdef DEBUG		
		peer_print(peer, peer_buffer, sizeof(peer_buffer));
		local_debug("cache_add[%d]: %s\n", filter->cache_curr_size, peer_buffer);
#endif	
		update_stat_ddd_filter(filter, 0);
	}
	return peer;
}
/* -------------------------------------------------------------------------- */
#if OLD_KERNEL_HOOK_4

#	if OLD_KERNEL_HOOK_3
static unsigned int nf_ddd_udp_fn(unsigned int hooknum, struct sk_buff *skb,
									const struct net_device *in, const struct net_device *out,
									int (*okfn)(struct sk_buff *))
#	else
static unsigned int nf_ddd_udp_fn(const struct nf_hook_ops *ops, struct sk_buff *skb,
									const struct net_device *in, const struct net_device *out,
									int (*okfn)(struct sk_buff *))
#	endif

#else

#if OLD_KERNEL_HOOK_4_4

static unsigned int nf_ddd_udp_fn(const struct nf_hook_ops *ops,
									struct sk_buff *skb,
									const struct nf_hook_state *state)
#	else
static unsigned int nf_ddd_udp_fn(void *priv,
									struct sk_buff *skb,
									const struct nf_hook_state *state)
#	endif

#endif
{
	struct iphdr 		*iph;
	struct udphdr 		*udph;
	struct tcphdr 		*tcph;
	struct ddd_peer		*peer;
	struct ip_packet	*ipp;
	struct timeval		tvstamp;
	struct ddd_filter	*filter;
	long				packet_time;
	int					filter_pkt;
	
	filter = find_ddd_filter(ddd_filters, ARRAY_SIZE(ddd_filters), skb->dev->name);
	
	if (!filter)
	{
		return NF_ACCEPT;
	}

	if (filter->update_ddd_force)
	{
		filter->update_ddd_force = 0;
		
		spin_lock(&filter->lock);
				
		update_stat_ddd_filter(filter, 1);
		
		spin_unlock(&filter->lock);		
	}
	
	if (!filter->enable) 
	{	
		filter->total_pkts 	= 0;
		filter->dup_pkts 	= 0;
		return NF_ACCEPT;
	}
#if OLD_KERNEL
	iph 	= (struct iphdr *)skb_network_header(skb);
	udph 	= (struct udphdr *)(skb_transport_header(skb)+sizeof(struct iphdr));
	tcph 	= (struct tcphdr *)(skb_transport_header(skb)+sizeof(struct iphdr));
#else
	iph		= ip_hdr(skb);
	udph	= udp_hdr(skb);
	tcph	= tcp_hdr(skb);
#endif
	
	ddd_gettimeofday(&tvstamp);
	
	filter->history_curr_max_size 	= filter->history_max_size;
	filter->cache_curr_max_size		= filter->cache_max_size;
	
	filter_pkt = (filter->filter_df == FILTER_DF_ALL) ||
				((filter->filter_df == FILTER_DF_ONLY_FLAG_OFF) && ((iph->frag_off & htons(IP_DF)) == 0)) ||
				((filter->filter_df == FILTER_DF_ONLY_FLAG_ON) && ((iph->frag_off & htons(IP_DF)) != 0));

	if (((iph->protocol == IPPROTO_UDP) || (iph->protocol == IPPROTO_TCP) || (iph->protocol == IPPROTO_ICMP)) &&
		filter_pkt
		)
	{
		spin_lock(&filter->lock);
		
		peer = NULL;
		if (iph->protocol == IPPROTO_ICMP)
		{
			if ((peer = cache_find(filter, iph->saddr, iph->daddr, 0, 0)) == NULL)
			{
				local_debug("cache_add");
				peer = cache_add(filter, iph->saddr, iph->daddr, 0, 0);
			}
		}
		else if (iph->protocol == IPPROTO_UDP)		
		{
			if ((peer = cache_find(filter, iph->saddr, iph->daddr, udph->source, udph->dest)) == NULL)
			{
				local_debug("cache_add");
				peer = cache_add(filter, iph->saddr, iph->daddr, udph->source, udph->dest);
			}
		}
		else if (iph->protocol == IPPROTO_TCP)		
		{
			if ((peer = cache_find(filter, iph->saddr, iph->daddr, tcph->source, tcph->dest)) == NULL)
			{
				local_debug("cache_add");
				peer = cache_add(filter, iph->saddr, iph->daddr, tcph->source, tcph->dest);
			}
		}

		if (peer)
		{
			peer->total_pkts++;
			peer->filterable |= (iph->id != 0);
			if (peer->filterable)
			{
				if ((ipp = history_find(filter, peer, iph->id)) != NULL)
				{
					/* find dup packet */
					packet_time = tvdiff(&tvstamp, &ipp->tstamp);
					if (packet_time > filter->ip_packet_ttl)
					{
						local_debug("old packet %ld\n", packet_time);
						ipp->tstamp = tvstamp;
					}
					else
					{
						local_debug("ddd_hook dup packet droped \n");
						peer->dup_pkts++;
						filter->dup_pkts++;
						filter->total_pkts++;

						spin_unlock(&filter->lock);

						return NF_DROP;
					}
				}
				else
				{
					local_debug("history_add\n");

					history_add(filter, peer, iph->id, &tvstamp);
				}
			}
			else
			{
				local_debug("ddd_hook accept packet for unfilterable peer\n");
				filter->total_pkts++;

				spin_unlock(&filter->lock);

				return NF_ACCEPT;
			}
		}

		filter->total_pkts++;
		
		update_stat_ddd_filter(filter, 0);
		
		spin_unlock(&filter->lock);
	}
	
	local_debug("ddd_hook packet accepted\n");	
	return NF_ACCEPT;
}
/* -------------------------------------------------------------------------- */
static struct nf_hook_ops nf_ddd_udp_hook_ops __read_mostly = 
{
	.pf 		= NFPROTO_IPV4,
	.priority 	= NF_IP_PRI_FILTER+1,
	.hooknum 	= NF_INET_LOCAL_IN,
#if OLD_KERNEL_HOOK_4_4
	.hook 		= nf_ddd_udp_fn,
	.owner		= THIS_MODULE
#else
	.hook 		= nf_ddd_udp_fn
#endif
};

/*
	new values
	.priority 	= NF_IP_PRI_FIRST,
	.hooknum 	= NF_INET_PRE_ROUTING,

	old values
	.priority 	= NF_IP_PRI_FILTER+1,
	.hooknum 	= NF_INET_LOCAL_IN,
*/


#ifdef CONFIG_DDD_UDP_SYSCTL

static int peer_ttl_min 		= 5000;
static int ip_packet_ttl_min 	= 1000;
static int cache_max_size_min	= 4;
static int history_max_size_min	= 2;

static int enable_min			= 0;
static int enable_max			= 1;

static int filter_df_min			= FILTER_DF_ONLY_FLAG_OFF;
static int filter_df_max			= FILTER_DF_ALL;

/* -------------------------------------------------------------------------- */
static int ddd_udp_ifname_handler(struct ctl_table *table, int write,
				void __user *buffer,
				size_t *lenp, loff_t *ppos)
{
	int ret = 0;
	ret = proc_dostring(table, write, buffer, lenp, ppos);
	
	
	return 0;
}
/* -------------------------------------------------------------------------- */
static int ddd_udp_minmax_handler(struct ctl_table *table, int write,
				void __user *buffer,
				size_t *lenp, loff_t *ppos)
{
	struct ddd_filter	*filter;
	int 				i;
	
	for (i=0;i<ARRAY_SIZE(ddd_filters);i++)
	{
		filter = &ddd_filters[i];
		filter->update_ddd_force = 1;
	}
	
	return proc_dointvec_minmax(table, write, buffer, lenp, ppos);
}
/* -------------------------------------------------------------------------- */
static struct ctl_table ddd_udp_table[] = {
	{
		.procname		= "ddd_udp",
		.data			= NULL,
		.maxlen			= sizeof(int),
		.mode			= 0644,
		.proc_handler	= ddd_udp_minmax_handler,
		.extra1			= &enable_min,
		.extra2			= &enable_max,
	},	
	{
		.procname		= "ddd_udp_ifname",
		.data			= NULL,
		.maxlen			= IFNAME_SIZE,
		.mode			= 0644,
		.proc_handler	= ddd_udp_ifname_handler
	},
	{
		.procname		= "ddd_peer_ttl",
		.data			= NULL,
		.maxlen			= sizeof(int),
		.mode			= 0644,
		.proc_handler	= ddd_udp_minmax_handler,
		.extra1			= &peer_ttl_min,
	},
	{
		.procname		= "ddd_packet_ttl",
		.data			= NULL,
		.maxlen			= sizeof(int),
		.mode			= 0644,
		.proc_handler	= ddd_udp_minmax_handler,
		.extra1			= &ip_packet_ttl_min,
	},
	{
		.procname		= "ddd_cache_max",
		.data			= NULL,
		.maxlen			= sizeof(int),
		.mode			= 0644,
		.proc_handler	= ddd_udp_minmax_handler,
		.extra1			= &cache_max_size_min,
	},	
	{
		.procname		= "ddd_history_max",
		.data			= NULL,
		.maxlen			= sizeof(int),
		.mode			= 0644,
		.proc_handler	= ddd_udp_minmax_handler,
		.extra1			= &history_max_size_min,
	},	
	{
		.procname		= "ddd_filter_df",
		.data			= NULL,
		.maxlen			= sizeof(int),
		.mode			= 0644,
		.proc_handler	= ddd_udp_minmax_handler,
		.extra1			= &filter_df_min,
		.extra2			= &filter_df_max,
	},	
	{}	
};

#if OLD_KERNEL_SYSCTL

static int init_sysctl_fs(void *net)
{
	
	struct ddd_filter	*filter;
	struct ctl_table 	*table = NULL;
	int 				i;

	if (net)
	{
		for (i=0;i<ARRAY_SIZE(ddd_filters);i++)
		{
			filter = &ddd_filters[i];
			
			if (!filter->table)
			{
				table = kmemdup(ddd_udp_table, sizeof(ddd_udp_table), GFP_ATOMIC);
				if (!table)
					goto out;

				table[0].data = &filter->enable;
				table[1].data = &filter->ifname;
				table[2].data = &filter->peer_ttl;
				table[3].data = &filter->ip_packet_ttl;
				table[4].data = &filter->cache_max_size;
				table[5].data = &filter->history_max_size;
				table[6].data = &filter->filter_df;
				
				

				filter->table		 = table;
				filter->table_header = register_sysctl_paths(filter->ddd_udp_ctl_path, table);
				if (!filter->table_header) 
				{
					printk(KERN_ERR "nf_ddd_udp[%d]: can't register to sysctl.\n", i);
					kfree(filter->table);
					filter->table = NULL;
				}
				else
				{
					printk(KERN_INFO "/proc/sys/net/ipv4/ddd_udp/0 created\n");
				}
			}
		}
	}	

	return 0;
out:
	printk(KERN_ERR "nf_ddd_udp[%d]: can't allocate memory.\n", i);
	return -ENOMEM;		
}
/* -------------------------------------------------------------------------- */
static int cleanup_sysctl_fs(void *net)
{
	int i;
	struct ddd_filter		*filter;
	
	for (i=0;i<ARRAY_SIZE(ddd_filters);i++)
	{
		filter = &ddd_filters[i];

		if (filter->table_header)
		{
			unregister_sysctl_table(filter->table_header);
			filter->table_header = NULL;
		}
		if (filter->table)
		{
			kfree(filter->table);
			filter->table = NULL;
		}
	}	
	
	return 0;	
}

#else /* OLD_KERNEL */
/* -------------------------------------------------------------------------- */
static int init_sysctl_fs(struct net *net)
{
	struct ddd_filter	*filter;
	char				sysctl_name[256];
	struct ctl_table 	*table = NULL;
	int 				i;

	if (net)
	{
		for (i=0;i<ARRAY_SIZE(ddd_filters);i++)
		{
			filter = &ddd_filters[i];
			
			if (!filter->table)
			{
				table = kmemdup(ddd_udp_table, sizeof(ddd_udp_table), GFP_ATOMIC);
				if (!table)
					goto out;

				table[0].data = &filter->enable;
				table[1].data = &filter->ifname;
				table[2].data = &filter->peer_ttl;
				table[3].data = &filter->ip_packet_ttl;
				table[4].data = &filter->cache_max_size;
				table[5].data = &filter->history_max_size;
				table[6].data = &filter->filter_df;
				
				memset(sysctl_name, 0, sizeof(sysctl_name));
				sprintf(sysctl_name, "net/ipv4/ddd_udp/%d", i);

				filter->table		 = table;
				filter->table_header = register_net_sysctl(net, sysctl_name, table);
				if (!filter->table_header) 
				{
					printk(KERN_ERR "nf_ddd_udp[%d]: can't register to sysctl.\n", i);
					kfree(filter->table);
					filter->table = NULL;
				}
				else
				{
					printk(KERN_INFO "/proc/sys/%s created\n", sysctl_name);
				}
			}
		}
	}	

	return 0;
out:
	printk(KERN_ERR "nf_ddd_udp[%d]: can't allocate memory.\n", i);
	return -ENOMEM;	
}
/* -------------------------------------------------------------------------- */
static int cleanup_sysctl_fs(struct net *net)
{
	int i;
	struct ddd_filter	*filter;
	
	for (i=0;i<ARRAY_SIZE(ddd_filters);i++)
	{
		local_debug("nf_ddd_udp cleanup_sysctl_fs.\n");		
		filter = &ddd_filters[i];

		if (filter->table_header)
		{
			unregister_net_sysctl_table(filter->table_header);
			filter->table_header = NULL;
		}
		if (filter->table)
		{
			kfree(filter->table);
			filter->table = NULL;
		}
	}	
	
	return 0;
}
/* -------------------------------------------------------------------------- */


#endif /* OLD_KERNEL */

#endif /* CONFIG_DDD_UDP_SYSCTL */


#ifdef CONFIG_DDD_UDP_PROC_FS

#define ddd_udp_procfs_file_name "ddd_udp"

int ddd_udp_seq_printf(struct seq_file *seq, struct ddd_filter *filter)
{
	if (seq && filter)
    {
		seq_printf(seq, filter->ddd_stat);
	}
 
    return 0;
}
/* -------------------------------------------------------------------------- */

/* -------------------------------------------------------------------------- */
static void *ddd_udp_seq_start(struct seq_file *seq, loff_t *pos)
{
	if (pos)
	{
		if (*pos >= ARRAY_SIZE(ddd_filters))
			return NULL;
	}
	return pos;
}
/* -------------------------------------------------------------------------- */
static void *ddd_udp_seq_next(struct seq_file *s, void *v, loff_t *pos)
{
	if (pos)
	{
		(*pos)++;
		if (*pos >= ARRAY_SIZE(ddd_filters))
			return NULL;
	}
	return pos;
}
/* -------------------------------------------------------------------------- */
static void ddd_udp_seq_stop(struct seq_file *s, void *v)
{
	/**/		
}
/* -------------------------------------------------------------------------- */
static int ddd_udp_seq_show(struct seq_file *seq, void *v)
{
	loff_t *pos = v;
	
	if ((0 <= *pos) && (*pos < ARRAY_SIZE(ddd_filters)))
	{
		ddd_udp_seq_printf(seq, &ddd_filters[*pos]);
	}
	
	return 0;
}
/* -------------------------------------------------------------------------- */
static const struct seq_operations ddd_udp_ops = {
	.start	= ddd_udp_seq_start,
	.next	= ddd_udp_seq_next,
	.stop	= ddd_udp_seq_stop,
	.show	= ddd_udp_seq_show
};

#if OLD_KERNEL_HOOK_4_18
/* -------------------------------------------------------------------------- */
static int ddd_udp_seq_open(struct inode *inode, struct file *file)
{
	return seq_open_net(inode, file, &ddd_udp_ops,
			    sizeof(struct ddd_filter));
}
/* -------------------------------------------------------------------------- */
static const struct file_operations ddd_udp_seq_fops = {
	.owner			= THIS_MODULE,
	.open           = ddd_udp_seq_open,
	.read           = seq_read,
	.llseek         = seq_lseek,
	.release		= seq_release_net,
};
/* -------------------------------------------------------------------------- */
static int init_proc_fs(struct net *net)
{
	if (net)
	{
		if (!proc_create(ddd_udp_procfs_file_name, S_IRUGO, net->proc_net, &ddd_udp_seq_fops))
			return -ENOMEM;
		printk(KERN_INFO "/proc/net/%s created\n", ddd_udp_procfs_file_name);
	}

	return 0;
}

#else /* OLD_KERNEL_HOOK_4_18 */

static int init_proc_fs(struct net *net)
{
	if (net)
	{
		if (!proc_create_net_data(ddd_udp_procfs_file_name, S_IRUGO, net->proc_net, &ddd_udp_ops,
				sizeof(struct ddd_filter),	(void *)ddd_filters))
			return -ENOMEM;
		printk(KERN_INFO "/proc/net/%s created\n", ddd_udp_procfs_file_name);
	}
		
	return 0;
}
#endif /* OLD_KERNEL_HOOK_4_18 */

/* -------------------------------------------------------------------------- */
static void cleanup_proc_fs(struct net *net)
{
	if (net)
	{
		remove_proc_entry(ddd_udp_procfs_file_name, net->proc_net);
		printk(KERN_INFO "/proc/net/%s removed\n", ddd_udp_procfs_file_name);
	}
}


#endif /* CONFIG_DDD_UDP_PROC_FS */
	
/* -------------------------------------------------------------------------- */


/* -------------------------------------------------------------------------- */
static int __net_init ddd_udp_net_init(struct net *net)
{
#ifdef CONFIG_DDD_UDP_SYSCTL
	init_sysctl_fs(net);	
#endif

#ifdef CONFIG_DDD_UDP_PROC_FS
	init_proc_fs(net);
#endif

#if !OLD_KERNEL_HOOK_4_13
	nf_register_net_hook(net, &nf_ddd_udp_hook_ops);
#endif
	return 0;
}
/* -------------------------------------------------------------------------- */
static void __net_exit ddd_udp_net_exit(struct net *net)
{
	local_debug("nf_ddd_udp ddd_udp_net_exit.\n");
#ifdef CONFIG_DDD_UDP_SYSCTL
	cleanup_sysctl_fs(net);	
#endif
	
#ifdef CONFIG_DDD_UDP_PROC_FS
	cleanup_proc_fs(net);
#endif

#if !OLD_KERNEL_HOOK_4_13
	nf_unregister_net_hook(net, &nf_ddd_udp_hook_ops);
#endif
}
/* -------------------------------------------------------------------------- */
static struct pernet_operations ddd_udp_net_ops = {
	.init = ddd_udp_net_init,
	.exit = ddd_udp_net_exit,
};
/* -------------------------------------------------------------------------- */
static int __init nf_ddd_udp_init(void)
{
	int i;
	printk(KERN_INFO "%s, Version %s.\n", DDD_UDP_DESCRIPTION, DDD_UDP_VERSION);
		
	for (i=0;i<ARRAY_SIZE(ddd_filters);i++)
	{
		init_ddd_udp(&ddd_filters[i]);
	}

	register_pernet_subsys(&ddd_udp_net_ops);
	
#if OLD_KERNEL_HOOK_4_13
	nf_register_hook(&nf_ddd_udp_hook_ops);
#endif
	return 0;
}
/* -------------------------------------------------------------------------- */
static void __exit nf_ddd_udp_exit(void)
{
	int i;

#if OLD_KERNEL_HOOK_4_13
	nf_unregister_hook(&nf_ddd_udp_hook_ops);
#endif
	
	unregister_pernet_subsys(&ddd_udp_net_ops);

	local_debug("nf_ddd_udp uninstalled\n");

	for (i=0;i<ARRAY_SIZE(ddd_filters);i++)
	{
		close_ddd_udp(&ddd_filters[i]);
	}
}
/* -------------------------------------------------------------------------- */


module_init(nf_ddd_udp_init);
module_exit(nf_ddd_udp_exit);


MODULE_LICENSE(DDD_UDP_LICENSE);
MODULE_AUTHOR(DDD_UDP_AUTHOR);
MODULE_DESCRIPTION(DDD_UDP_DESCRIPTION);
MODULE_VERSION(DDD_UDP_VERSION);

